from collections import OrderedDict


from easydict import EasyDict as edict
import numpy as np
from pyvisa import ResourceManager
from qtpy.QtCore import Signal

from pymodaq.utils.daq_utils import ThreadCommand, getLineInfo
from pymodaq.utils.data import  DataFromPlugins, DataToExport
from pymodaq.control_modules.viewer_utility_classes import DAQ_Viewer_base, main, comon_parameters
from pymodaq.utils.enums import BaseEnum



VISA_rm = ResourceManager()
COM_PORTS = []
for name, rinfo in VISA_rm.list_resources_info().items():
    if rinfo.alias is not None:
        COM_PORTS.append(rinfo.alias)
    else:
        COM_PORTS.append(name)

class DAQ_0DViewer_Keithley_Pico_type(BaseEnum):
    """
        Enum class of Keithley_Pico_type

        =============== =========
        **Attributes**  **Type**
        *Pico_648X*     int
        *Pico_6430*     int
        *Pico_6514*     int
        =============== =========
    """
    Pico_648X = 0
    Pico_6430 = 1
    Pico_6514 = 2


class DAQ_0DViewer_Keithley_Pico(DAQ_Viewer_base):
    """
        ==================== ========================
        **Attributes**        **Type**
        *data_grabed_signal*  instance of Signal
        *VISA_rm*             ResourceManager
        *com_ports*           
        *params*              dictionnary list
        *keithley*
        *settings*
        ==================== ========================
    """
    data_grabed_signal = Signal(list)

    ##checking VISA ressources

#    import serial.tools.list_ports;
#    com_ports=[comport.device for comport in serial.tools.list_ports.comports()]

    params = comon_parameters + [
        {'title': 'VISA:', 'name': 'VISA_ressources', 'type': 'list', 'limits': COM_PORTS},
        {'title': 'Keithley Type:', 'name': 'keithley_type', 'type': 'list',
         'limits': DAQ_0DViewer_Keithley_Pico_type.names()},
        {'title': 'Id:', 'name': 'id', 'type': 'text', 'value': ""},
        {'title': 'Timeout (ms):', 'name': 'timeout', 'type': 'int', 'value': 10000, 'default': 10000, 'min': 2000},
        {'title': 'Configuration:', 'name': 'config', 'type': 'group', 'children': [
            {'title': 'Meas. type:', 'name': 'meas_type', 'type': 'list', 'value': 'CURR', 'default': 'CURR',
             'limits': ['CURR', 'VOLT', 'RES', 'CHAR']},

        ]},
    ]

    def ini_attributes(self):
        pass

    def ini_detector(self, controller=None):
        """
            Initialisation procedure of the detector.

            Returns
            -------

                The initialized status.

            See Also
            --------
            daq_utils.ThreadCommand
        """

        self.controller = self.ini_detector_init(old_controller=controller,
                                                 new_controller=
                                                 VISA_rm.open_resource(
                                                     self.settings['VISA_ressources'], read_termination='\r'))

        self.controller.timeout = self.settings['timeout']

        self.controller.write("*rst; status:preset; *cls;")
        txt = self.controller.query('*IDN?')
        self.settings.child(('id')).setValue(txt)
        self.controller.write('CONF:' + self.settings.child('config', 'meas_type').value())
        self.controller.write(':FORM:ELEM READ;DATA ASC;')
        self.controller.write('ARM:SOUR IMM;')
        self.controller.write('ARM:COUNt 1;')
        self.controller.write('TRIG:SOUR IMM;')
        # %%
        data = self.controller.query_ascii_values('READ?')

        self.status.initialized = True
        self.status.controller = self.controller
        return self.status


    def commit_settings(self, param):
        """
            Activate the parameters changes in the hardware.

            =============== ================================= ============================
            **Parameters**   **Type**                         **Description**
            *param*         instance of pyqtgraph.parameter   The parameter to be checked.
            =============== ================================= ============================

            See Also
            --------
            daq_utils.ThreadCommand
        """
        try:
            if param.name() == 'timeout':
                self.controller.timeout = self.settings.child(('timeout')).value()
            elif param.name() == 'meas_type':
                self.controller.write('CONF:' + param.value())


        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [getLineInfo() + str(e), 'log']))

    def close(self):
        """
            close the current instance of Keithley viewer.
        """
        if self.controller is not None:
            self.controller.close()

    def grab_data(self, Naverage=1, **kwargs):
        """
            | Start new acquisition.
            | grab the current values with keithley profile procedure.
            | Send the data_grabed_signal once done.

            =============== ======== ===============================================
            **Parameters**  **Type**  **Description**
            *Naverage*      int       Number of values to average
            =============== ======== ===============================================
        """
        data_tot = []
        self.controller.write('ARM:SOUR IMM;')
        self.controller.write('ARM:COUNt 1;')
        self.controller.write('TRIG:SOUR IMM;')
        self.controller.write('TRIG:COUN {:};'.format(Naverage))
        data_tot = self.controller.query_ascii_values('READ?')
        # for ind in range(Naverage):
        #    data_tot.append(self.controller.query_ascii_values('READ?')[0])
        dwa = DataFromPlugins(name='Keithley', data=[np.array([np.mean(np.array(data_tot))])])
        self.dte_signal.emit(DataToExport('Keithley', data=[dwa]))

    def stop(self):
        """
            not implemented?
        """
        return ""


if __name__ == '__main__':
    main(__file__, init=False)



import pyvisa as visa


class Keithley2110VISADriver:
    """
        VISA class driver for the Keithley 2110  5 1/2 Digit DMM
        This class relies on pyvisa module to communicate with the instrument via VISA protocol
        Please refer to the instrument reference manual available at:
        https://download.tek.com/manual/2110-901-01(C-Aug2013)(Ref).pdf
    """
    def __init__(self, rsrc_name, pyvisa_backend='@ni'):
        """
        Parameters
        ----------
        rsrc_name   (string)        VISA Resource name
        pyvisa_backend  (string)    Expects a pyvisa backend identifier or a path to the visa backend dll (ref. to pyvisa)
        """
        rm = visa.highlevel.ResourceManager(pyvisa_backend)
        self._instr = rm.open_resource(rsrc_name)

        self._instr.read_termination = '\n'
        self._instr.write_termination = '\n'

    def close(self):
        self._instr.close()

    def get_identification(self):
        self._instr.query("*IDN?")

    def reset(self):
        self._instr.write("*CLS")
        self._instr.write("*RST")

    def read(self):
        return float(self._instr.query("READ?"))

    def set_mode(self, mode, **kwargs):
        """

        Parameters
        ----------
        mode    (string)    Measurement configuration ('VDC', 'VAC', 'IDC', 'IAC', 'R2W' and 'R4W' modes are supported)
        kwargs  (dict)      Used to pass optional arguments ('range' and 'resolution' are the only supported keys)

        Returns
        -------

        """
        assert (isinstance(mode, str))
        mode = mode.lower()

        cmd = ':CONF:'

        if mode == "Ohm2".lower() or mode == "R2W".lower():
            cmd += "RES"
        elif mode == "Ohm4".lower() or mode == "R4W".lower():
            cmd += "FRES"
        elif mode == "VDC".lower() or mode == "V".lower():
            cmd += "VOLT:DC"
        elif mode == "VAC".lower():
            cmd += "VOLT:AC"
        elif mode == "IDC".lower() or mode == "I".lower():
            cmd += "CURR:DC"
        elif mode == "IAC".lower():
            cmd += "CURR:AC"

        if 'range' in kwargs.keys():
            cmd += ' ' + str(kwargs['range'])
            if 'resolution' in kwargs.keys():
                cmd += ',' + str(kwargs['resolution'])
        elif 'resolution' in kwargs.keys():
            cmd += ' DEF,' + str(kwargs['resolution'])

        self._instr.write(cmd)


if __name__ == "__main__":
    try:
        k2110 = Keithley2110VISADriver("K2110")
        k2110.reset()
        k2110.set_mode('Ohm2')
        k2110.set_mode('R4W', range=10, resolution='MAX')
        k2110.set_mode('R4W', resolution='MIN')
        k2110.set_mode('IAC', range=0.001, resolution='MIN')
        k2110.set_mode('vdc', range=0.1, resolution='0.0001')

        print(k2110.read())

        k2110.close()

    except Exception as e:
        print("Exception ({}): {}".format(type(e), str(e)))



import numpy as np
from pymodaq.utils.daq_utils import ThreadCommand
from pymodaq.utils.data import DataFromPlugins, DataToExport
from pymodaq.control_modules.viewer_utility_classes import DAQ_Viewer_base, comon_parameters, main
from pymodaq.utils.parameter import Parameter
from pymodaq_plugins_keithley import config
from pymodaq_plugins_keithley.hardware.keithley27XX.keithley27XX_VISADriver import Keithley27XXVISADriver as Keithley
from pymodaq.utils.logger import set_logger, get_module_name
logger = set_logger(get_module_name(__file__))


class DAQ_0DViewer_Keithley27XX(DAQ_Viewer_base):
    """ Keithley plugin class for a OD viewer.

    This object inherits all functionalities to communicate with PyMoDAQ’s DAQ_Viewer module through inheritance via
    DAQ_Viewer_base. It makes a bridge between the DAQ_Viewer module and the keithley27XX_VISADriver.

    :param controller: The particular object that allow the communication with the keithley27XX_VISADriver.
    :type  controller:  object

    :param params: Parameters displayed in the daq_viewer interface
    :type params: dictionary list
    """
    rsrc_name: str
    instr: str
    panel: str
    channels_in_selected_mode: str
    resources_list = []
    
    # Read configuration file
    for instr in config["Keithley", "27XX"].keys():
        if "INSTRUMENT" in instr:
            resources_list += [config["Keithley", "27XX", instr, "rsrc_name"]]
    logger.info("resources list = {}" .format(resources_list))

    params = comon_parameters + [
        {'title': 'Resources', 'name': 'resources', 'type': 'list', 'limits': resources_list,
         'value': resources_list[0]},
        {'title': 'Keithley', 'name': 'Keithley_Params', 'type': 'group', 'children': [
            {'title': 'Panel', 'name': 'panel', 'type': 'list', 'limits': ['select panel to use', 'FRONT', 'REAR'],
             'value': 'select panel to use'},
            {'title': 'ID', 'name': 'ID', 'type': 'text', 'value': ''},
            {'title': 'FRONT panel', 'name': 'frontpanel', 'visible': False, 'type': 'group', 'children': [
                {'title': 'Mode', 'name': 'frontmode', 'type': 'list',
                 'limits': ['VOLT:DC', 'VOLT:AC', 'CURR:DC', 'CURR:AC', 'RES', 'FRES', 'FREQ', 'TEMP'],
                 'value': 'VOLT:DC'},
            ]},
            {'title': 'REAR panel', 'name': 'rearpanel', 'visible': False, 'type': 'group', 'children': [
                {'title': 'Mode', 'name': 'rearmode', 'type': 'list',
                 'limits': ['SCAN_LIST', 'VOLT:DC', 'VOLT:AC', 'CURR:DC', 'CURR:AC', 'RES', 'FRES', 'FREQ', 'TEMP'],
                 'value': 'SCAN_LIST'}
            ]},
        ]},
    ]

    def __init__(self, parent=None, params_state=None):
        super().__init__(parent, params_state)

    def ini_attributes(self):
        """Attributes init when DAQ_0DViewer_Keithley class is instanced"""
        self.controller: Keithley = None
        self.channels_in_selected_mode = None
        self.rsrc_name = None
        self.panel = None
        self.instr = None

    def commit_settings(self, param: Parameter):
        """Apply the consequences of a change of value in the detector settings"""
        if param.name() == 'panel':
            for limit in ['REAR', 'FRONT']:
                if not limit == param.value():
                    if param.value() == 'select panel to use':
                        self.settings.child('Keithley_Params', 'frontpanel').show()
                        self.settings.child('Keithley_Params', 'rearpanel').show()
                    else:
                        self.settings.child('Keithley_Params', param.value().lower() + param.name()).show()
                        self.settings.child('Keithley_Params', limit.lower() + param.name()).hide()
        if 'mode' in param.name():
            """Updates the newly selected measurement mode"""
            # Read the configuration file to determine which mode to use and send corresponding instruction to driver
            if self.panel == 'FRONT':
                value = param.value()
                self.controller.set_mode(value)
            elif self.panel == 'REAR':
                value = 'SCAN_' + param.value()
                self.channels_in_selected_mode = self.controller.set_mode(value)
            current_error = self.controller.get_error()
            if current_error != '0,"No error"':
                logger.error("The following error has been raised by the Keithley:\
                        {} => Please refer to the User Manual to correct it\n\
                        Note: To make sure channels are well configured in the .toml file,\
                        refer to section 15 'SCPI Reference Tables', Table 15-5" .format(current_error))
        if 'CURR' in param.value():
            """Verify if the switching modules support current measurement"""
            if self.controller.non_amp_module["MODULE01"] and self.controller.non_amp_module["MODULE02"]:
                logger.info("Both modules don't support current measurement")
            if self.controller.non_amp_module["MODULE01"] and not self.controller.non_amp_module["MODULE02"]:
                logger.info("Both modules don't support current measurement")
            if self.controller.non_amp_module["MODULE02"] and not self.controller.non_amp_module["MODULE01"]:
                logger.info("Both modules don't support current measurement")

    def ini_detector(self, controller=None):
        """Detector communication initialization

        :param controller: Custom object of a PyMoDAQ plugin (Slave case). None if one actuator/detector by controller.
        :type controller: object

        :return: Initialization status, false if it failed otherwise True
        :rtype: bool
        """
        logger.info("Detector 0D initialized")

        if self.settings.child('controller_status').value() == "Slave":
            if controller is None:
                raise Exception('no controller has been defined externally while this detector is a slave one')
            else:
                self.controller = controller
        else:
            try:
                # Select the resource to connect with and load the dedicated configuration
                for instr in config["Keithley", "27XX"]:
                    if "INSTRUMENT" in instr:
                        if config["Keithley", "27XX", instr, "rsrc_name"] == self.settings["resources"]:
                            self.rsrc_name = config["Keithley", "27XX", instr, "rsrc_name"]
                            self.panel = config["Keithley", "27XX", instr, "panel"].upper()
                            self.instr = instr
                            logger.info("Panel configuration 0D_viewer: {}" .format(self.panel))
                assert self.rsrc_name is not None, "rsrc_name"
                assert self.panel is not None, "panel"
                self.controller = Keithley(self.rsrc_name)
            except AssertionError as err:
                logger.error("{}: {} did not match any configuration".format(type(err), str(err)))
            except Exception as e:
                raise Exception('No controller could be defined because an error occurred \
                while connecting to the instrument. Error: {}'.format(str(e)))

        # Keithley initialization & identification
        self.controller.init_hardware()
        txt = self.controller.get_idn()
        self.settings.child('Keithley_Params', 'ID').setValue(txt)

        # Initialize detector communication and set the default value (SCAN_LIST)
        if self.panel == 'FRONT':
            self.settings.child('Keithley_Params', 'rearpanel').visible = False
            value = self.settings.child('Keithley_Params', 'frontpanel', 'frontmode').value()
            self.controller.current_mode = value
            self.controller.set_mode(value)
        elif self.panel == 'REAR':
            self.settings.child('Keithley_Params', 'frontpanel').visible = False
            self.settings.child('Keithley_Params', 'frontpanel').value = 'REAR'
            self.controller.configuration_sequence()
            value = 'SCAN_' + self.settings.child('Keithley_Params', 'rearpanel', 'rearmode').value()
            self.channels_in_selected_mode = self.controller.set_mode(value)
            logger.info("Channels to plot : {}" .format(self.channels_in_selected_mode))
        logger.info("DAQ_viewer command sent to keithley visa driver : {}" .format(value))

        self.status.initialized = True
        self.status.controller = self.controller

        return self.status

    def close(self):
        """Terminate the communication protocol"""
        self.controller.close()
        logger.info("communication ended successfully")

    def grab_data(self, Naverage=1, **kwargs):
        """Start a grab from the detector

        :param Naverage: Number of hardware averaging (if hardware averaging is possible,
            self.hardware_averaging should be set to True in class preamble, and you should code this implementation)
        :type Naverage: int

        :param kwargs: others optionals arguments
        :type kwargs: dict
        """
        # ACQUISITION OF DATA
        if self.panel == 'FRONT':
            data_tot = self.controller.data()
            data_measurement = data_tot[1]
        elif self.panel == 'REAR':
            channels_in_selected_mode = self.channels_in_selected_mode[1:-1].replace('@', '')
            chan_to_plot = []
            data_tot = self.controller.data()
            data_measurement = data_tot[1]
            for i in range(len(channels_in_selected_mode.split(','))):
                chan_to_plot.append('Channel ' + str(channels_in_selected_mode.split(',')[i]))
            # Affect each value to the corresponding channel
            dict_chan_value = dict(zip(channels_in_selected_mode.split(','), data_measurement))
        # Dictionary linking channel's modes to physical quantities
        dict_label_mode = {'VOLT:DC': 'Voltage', 'VOLT:AC': 'Voltage', 'CURR:DC': 'Current', 'CURR:AC': 'Current',
                           'RES': 'Resistance', 'FRES': 'Resistance', 'FREQ': 'Frequency', 'TEMP': 'Temperature'}
        # EMISSION OF DATA
        # When reading the scan_list, data are displayed and exported grouped by mode
        if not self.controller.reading_scan_list:
            label = dict_label_mode[self.controller.current_mode]
            if self.panel == 'FRONT':
                labels = 'Front input'
            elif self.panel == 'REAR':
                labels = [chan_to_plot[i] for i in range(len(chan_to_plot))]
            dte = DataToExport(name='keithley',
                               data=[DataFromPlugins(name=label,
                                                     data=[np.array([data_measurement[i]]) for i in
                                                           range(len(data_measurement))],
                                                     dim='Data0D',
                                                     labels=labels)])

        # Reading only channels configured in the selected mode
        elif self.controller.reading_scan_list:
            dte = DataToExport(name='keithley',
                               data=[DataFromPlugins(name=dict_label_mode[key],
                                                     data=[np.array([dict_chan_value[str(chan)]]) for chan in
                                                           self.controller.modes_channels_dict.get(key)],
                                                     dim='Data0D',
                                                     labels=['Channel ' + str(chan) for chan in
                                                             self.controller.modes_channels_dict.get(key)]
                                                     ) for key in self.controller.modes_channels_dict.keys() if
                                     self.controller.modes_channels_dict.get(key) != []])
        self.dte_signal.emit(dte)

    def stop(self):
        """Stop the current grab hardware wise if necessary"""
        self.emit_status(ThreadCommand('Update_Status', ['Acquisition stopped']))
        return ''


if __name__ == '__main__':
    main(__file__)



from easydict import EasyDict as edict
from pymeasure.instruments.keithley import Keithley2400
from pymeasure.adapters import VISAAdapter, PrologixAdapter

from pymodaq.control_modules.move_utility_classes import DAQ_Move_base  # base class
from pymodaq.control_modules.move_utility_classes import comon_parameters, main  # common set of parameters for all actuators
from pymodaq.utils.daq_utils import ThreadCommand, getLineInfo
from pymodaq.utils.logger import set_logger, get_module_name  # object used to send info back to the main thread
from pymodaq.utils.parameter.utils import iter_children
from pyvisa import ResourceManager


logger = set_logger(get_module_name(__file__))

rm = ResourceManager()

VISA_RESSOURCES = rm.list_resources()
ADAPTERS = dict(VISA=VISAAdapter, Prologix=PrologixAdapter)
SOURCE_MODES = ['Current', 'Voltage']
EPSILON_CURRENT = 1e-5
EPSILON_VOLTAGE = 1e-3


class DAQ_Move_Keithley2400(DAQ_Move_base):
    """
        Wrapper object to access the Mock fonctionnalities, similar wrapper for all controllers.

        =============== ==============
        **Attributes**    **Type**
        *params*          dictionnary
        =============== ==============
    """
    _controller_units = 'A'
    is_multiaxes = False  # set to True if this plugin is controlled for a multiaxis controller (with a unique communication link)
    stage_names = []  # "list of strings of the multiaxes
    _epsilon = 1e-5

    params = [   {'title': 'Adapter:', 'name': 'adapter', 'type': 'list', 'limits': list(ADAPTERS.keys())},
                 {'title': 'VISA Ressources:', 'name': 'visa_ressource', 'type': 'list', 'limits': VISA_RESSOURCES},
                 {'title': 'Info:', 'name': 'info', 'type': 'str', 'value': '', 'readonly': True},
                 {'title': 'Source Mode:', 'name': 'source_mode', 'type': 'list', 'limits': SOURCE_MODES},
                 {'title': 'Enabled:', 'name': 'enabled', 'type': 'led_push', 'value': False},
                 {'title': 'Current Mode:', 'name': 'current_mode', 'type': 'group', 'children': [
                     {'title': 'Current Range:', 'name': 'current_range', 'type': 'float', 'value': 10e-3, 'min': 0.},
                     {'title': 'Compliance Voltage:', 'name': 'voltage_compliance', 'type': 'float',
                      'value': 10, 'min': 0.}]},
                 {'title': 'Voltage Mode:', 'name': 'voltage_mode', 'type': 'group', 'visible': True, 'children': [
                     {'title': 'Voltage Range:', 'name': 'voltage_range', 'type': 'float', 'value': 10, 'min': 0.,
                      'max': 210.},
                     {'title': 'Compliance Current:', 'name': 'current_compliance', 'type': 'float',
                      'value': 5e-1, 'min': 0.}]},

                 {'title': 'MultiAxes:', 'name': 'multiaxes', 'type': 'group', 'visible': is_multiaxes, 'children': [
                     {'title': 'is Multiaxes:', 'name': 'ismultiaxes', 'type': 'bool', 'value': is_multiaxes,
                      'default': False},
                     {'title': 'Status:', 'name': 'multi_status', 'type': 'list', 'value': 'Master',
                      'limits': ['Master', 'Slave']},
                     {'title': 'Axis:', 'name': 'axis', 'type': 'list', 'limits': stage_names},

                 ]}] + comon_parameters

    def __init__(self, parent=None, params_state=None):
        """
            Initialize the the class

            ============== ================================================ ==========================================================================================
            **Parameters**  **Type**                                         **Description**

            *parent*        Caller object of this plugin                    see DAQ_Move_main.DAQ_Move_stage
            *params_state*  list of dicts                                   saved state of the plugins parameters list
            ============== ================================================ ==========================================================================================

        """

        super().__init__(parent, params_state)
        self._enabled = False

    def check_position(self):
        """Get the current position from the hardware with scaling conversion.

        Returns
        -------
        float: The position obtained after scaling conversion.
        """
        # if self.enabled:
        #     if self.settings.child('source_mode').value() == 'Current':
        #         self.controller.measure_current()
        #         pos = self.controller.current
        #     else:
        #         self.controller.measure_voltage()
        #         pos = self.controller.voltage
        #     if isinstance(pos, list):
        #         logger.debug(f"Got multiple return values for {self.settings.child('source_mode').value()}"
        #                      f": {pos}")
        #         pos = pos[0]
        # else:
        #     pos = 0.
        pos = self.current_position  # bypass checking
        pos = self.get_position_with_scaling(pos)

        self.emit_status(ThreadCommand('check_position', [pos]))
        return pos

    def close(self):
        """
        Terminate the communication protocol
        """
        self.controller.shutdown()

    @property
    def enabled(self):
        return self._enabled

    def enable_source(self, enable=True):
        self._enabled = enable
        if enable:
            self.controller.enable_source()
        else:
            self.controller.disable_source()
        self.settings.child('enabled').setValue(enable)

    def set_source(self, source_mode='Current', range=None, compliance=None):
        if source_mode == 'Current':
            if compliance is None:
                compliance = 0.1
            self.controller.apply_current(current_range=range, compliance_voltage=compliance)
            self.settings.child('epsilon').setValue(EPSILON_CURRENT)
        else:
            if compliance is None:
                compliance = 10
            self.controller.apply_voltage(voltage_range=range, compliance_current=compliance)
            self.settings.child('epsilon').setValue(EPSILON_VOLTAGE)
    def get_range_compliance(self):
        if self.settings.child('source_mode').value() == 'Current':
            range = self.settings.child('current_mode', 'current_range').value()
            compliance = self.settings.child('current_mode', 'voltage_compliance').value()
        else:
            range = self.settings.child('voltage_mode', 'voltage_range').value()
            compliance = self.settings.child('voltage_mode', 'current_compliance').value()
        return range, compliance

    def commit_settings(self, param):
        if param.name() == "source_mode" or \
            param.name() in iter_children(self.settings.child('current_mode'), []) or \
                param.name() in iter_children(self.settings.child('voltage_mode'), []):

            if self.enabled:  # if was enabled, disabled it
                self.enable_source(False)
            self.set_source(param.value(), *self.get_range_compliance())
            if param.name() == "source_mode":
                self.settings.child('current_mode').show(param.value() == 'Current')
                self.settings.child('voltage_mode').show(param.value() != 'Current')
                if param.value() == 'Current':
                    self.controller.source_current = 0
                    self.controller.measure_voltage()
                else:
                    self.controller.source_voltage = 0.
                    self.controller.measure_current()

        elif param.name() == 'enabled':
            self.enable_source(param.value())

    def ini_stage(self, controller=None):
        """Actuator communication initialization

        Parameters
        ----------
        controller: (object) custom object of a PyMoDAQ plugin (Slave case). None if only one actuator by controller (Master case)

        Returns
        -------
        self.status (edict): with initialization status: three fields:
            * info (str)
            * controller (object) initialized controller
            *initialized: (bool): False if initialization failed otherwise True
        """
        try:
            # initialize the stage and its controller status
            # controller is an object that may be passed to other instances of DAQ_Move_Mock in case
            # of one controller controlling multiactuators (or detector)

            self.status.update(edict(info="", controller=None, initialized=False))

            # check whether this stage is controlled by a multiaxe controller (to be defined for each plugin)
            # if multiaxes then init the controller here if Master state otherwise use external controller
            if self.settings.child('multiaxes', 'ismultiaxes').value() and self.settings.child('multiaxes',
                                   'multi_status').value() == "Slave":
                if controller is None:
                    raise Exception('no controller has been defined externally while this axe is a slave one')
                else:
                    self.controller = controller
            else:  # Master stage

                adapter = \
                    ADAPTERS[self.settings.child('adapter').value()](self.settings.child('visa_ressource').value())
                self.controller = Keithley2400(adapter)  # when writing your own plugin replace this line
                self.commit_settings(self.settings.child('source_mode'))

                #####################################

            self.status.info = self.controller.id
            self.settings.child('info').setValue(self.status.info)
            self.status.controller = self.controller
            self.status.initialized = True
            return self.status

        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [getLineInfo() + str(e), 'log']))
            self.status.info = getLineInfo() + str(e)
            self.status.initialized = False
            return self.status

    def move_Abs(self, position):
        """ Move the actuator to the absolute target defined by position

        Parameters
        ----------
        position: (flaot) value of the absolute target positioning
        """

        position = self.check_bound(position)  #if user checked bounds, the defined bounds are applied here
        position = self.set_position_with_scaling(position)  # apply scaling if the user specified one
        if self.settings.child('source_mode').value() == 'Current':
            self.controller.source_current = position
        else:
            self.controller.source_voltage = position

        if self.enabled:
            if self.settings.child('source_mode').value() == 'Current':
                self.controller.measure_voltage()
                pos = self.controller.voltage
            else:
                self.controller.measure_current()
                pos = self.controller.current


        self.target_position = position
        self.current_position = self.target_position #bypass checking

    def move_Rel(self, position):
        """ Move the actuator to the relative target actuator value defined by position

        Parameters
        ----------
        position: (flaot) value of the relative target positioning
        """
        position = self.check_bound(self.current_position+position)-self.current_position
        self.target_position = position + self.current_position
        self.move_Abs(self.target_position)
        ##############################

    def move_Home(self):
        """
          Send the update status thread command.
            See Also
            --------
            daq_utils.ThreadCommand
        """

        self.move_Abs(0)

    def stop_motion(self):
        """
        Call the specific move_done function (depending on the hardware).

        See Also
        --------
        move_done
        """
        self.move_done()  # to let the interface know the actuator stopped


if __name__ == '__main__':
    main(__file__, init=False)



from pymodaq.control_modules.viewer_utility_classes import DAQ_Viewer_base, comon_parameters
from pymodaq.utils.data import DataFromPlugins
from easydict import EasyDict as edict
from collections import OrderedDict
from ...hardware.keithley2110.keithley2110_VISADriver import Keithley2110VISADriver as Keithley2110


class DAQ_0DViewer_Keithley2110(DAQ_Viewer_base):
    """
        Naive implementation of a DAQ 0D Viewer using the Keithley 2110 as data source
        This DAQ0D Viewer plugin only supports measurement mode selection and a simple data read acquisition mechanism
        with no averaging supported
        =============== =================
        **Attributes**  **Type**
        *params*        dictionnary list
        *x_axis*        1D numpy array
        *ind_data*      int
        =============== =================
    """
    params = comon_parameters+[
        {'title': 'Keithley2210 Parameters',  'name': 'K2110Params', 'type': 'group', 'children': [
            {'title': 'Mode', 'name': 'mode', 'type': 'list', 'limits': ['VDC', 'VAC', 'R2W', 'R4W'], 'value': 'VDC'}

        ]}
    ]

    def __init__(self, parent=None, params_state=None): # init_params is a list of tuple where each tuple contains info on a 1D channel (Ntps,amplitude, width, position and noise)
        super(DAQ_0DViewer_Keithley2110, self).__init__(parent, params_state)
        self.x_axis = None
        self.ind_data = 0


    def commit_settings(self, param):
        """
            ============== ========= =================
            **Parameters**  **Type**  **Description**
            *param*        child node  could be the following setting parameter: 'mode'
            ============== ========= =================
        """
        if param.name() == 'mode':
            """Updates the newly selected measurement mode"""
            self.controller.set_mode(param.value())

    def ini_detector(self, controller=None):
        """
            Initialisation procedure of the detector.

            Returns
            -------
                the initialized status.
        """

        self.status.update(edict(initialized=False, info="", x_axis=None, y_axis=None, controller=None))
        if self.settings.child(('controller_status')).value() == "Slave":
            if controller is None: 
                raise Exception('no controller has been defined externally while this detector is a slave one')
            else:
                self.controller = controller
        else:
            try:
                self.controller = Keithley2110('K2110')
            except Exception as e:
                raise Exception('No controller could be defined because an error occurred\
                 while connecting to the instrument. Error: {}'.format(str(e)))

        self.controller.set_mode(self.settings.child('K2110Params', 'mode').value())

        # initialize viewers with the future type of data
        self.data_grabed_signal.emit([DataFromPlugins(name='Keithley2110', data=[0], dim='Data0D', labels=['Meas', 'Time'])])

        self.status.initialized = True
        self.status.controller = self.controller
        return self.status

    def close(self):
        """
            not implemented.
        """
        pass

    def grab_data(self, Naverage=1, **kwargs):
        """
            | Start new acquisition.
            |
            |
            | Send the data_grabed_signal once done.

            =============== ======== ===============================================
            **Parameters**  **Type**  **Description**
            *Naverage*      int       specify the threshold of the mean calculation
            =============== ======== ===============================================

        """
        data = self.controller.read()
        self.data_grabed_signal.emit([utils.DataFromPlugins(name='K2110', data=[[data]], dim='Data0D',)])
        self.ind_data += 1

    def stop(self):
        """
            not implemented.
        """

        return ""



import numpy as np
import pyvisa as visa
from pymodaq_plugins_keithley import config
from pymodaq.utils.logger import set_logger, get_module_name
logger = set_logger(get_module_name(__file__))


class Keithley27XXVISADriver:
    """VISA class driver for the Keithley 27XX Multimeter/Switch System

    This class relies on pyvisa module to communicate with the instrument via VISA protocol.
    Please refer to the instrument reference manual available at:
    https://download.tek.com/manual/2700-900-01K_Feb_2016.pdf
    https://download.tek.com/manual/2701-900-01G_Feb_2016.pdf
    """
    # List the Keithley instruments the user has configured from the .toml configuration file
    list_instruments = {}
    for instr in config["Keithley", "27XX"].keys():
        if "INSTRUMENT" in instr:
            list_instruments[instr] = config["Keithley", "27XX", instr, "rsrc_name"]
    logger.info("Configured instruments: {}".format(list(list_instruments.items())))

    # Non-amps modules
    non_amp_module = {"MODULE01": False, "MODULE02": False}
    non_amp_modules_list = ['7701', '7703', '7706', '7707', '7708', '7709']

    # Channels & modes attributes
    channels_scan_list = ''
    modes_channels_dict = {'VOLT:DC': [],
                           'VOLT:AC': [],
                           'CURR:DC': [],
                           'CURR:AC': [],
                           'RES': [],
                           'FRES': [],
                           'FREQ': [],
                           'TEMP': []}
    sample_count_1 = False
    reading_scan_list = False
    current_mode = ''

    def __init__(self, rsrc_name):
        """Initialize KeithleyVISADriver class

        :param rsrc_name: VISA Resource name
        :type rsrc_name: string
        """
        self._instr = None
        self.rsrc_name = rsrc_name
        self.instr = ""
        self.configured_modules = {}

    def init_hardware(self, pyvisa_backend='@ivi'):
        """Initialize the selected VISA resource
        
        :param pyvisa_backend: Expects a pyvisa backend identifier or a path to the visa backend dll (ref. to pyvisa)
        :type pyvisa_backend: string
        """
        # Open connexion with instrument
        rm = visa.highlevel.ResourceManager(pyvisa_backend)
        logger.info("Resources detected by pyvisa: {}".format(rm.list_resources(query='?*')))
        try:
            self._instr = rm.open_resource(self.rsrc_name,
                                           write_termination="\n",
                                           read_termination="\n",
                                           )
            self._instr.timeout = 10000
            # Check if the selected resource match the loaded configuration
            model = self.get_idn()[32:36]
            if "27" not in model:
                logger.warning("Driver designed to use Keithley 27XX, not {} model. Problems may occur.".format(model))
            for instr in config["Keithley", "27XX"]:
                if type(config["Keithley", "27XX", instr]) == dict:
                    if self.rsrc_name in config["Keithley", "27XX", instr, "rsrc_name"]:
                        self.instr = instr
            logger.info("Instrument selected: {} ".format(config["Keithley", "27XX", self.instr, "rsrc_name"]))
            logger.info("Keithley model : {}".format(config["Keithley", "27XX", self.instr, "model_name"]))
            try:
                # Load the configuration matching the selected module
                cards = self.get_card().split(',')
                logger.info("card : {}".format(cards))
                try:
                    assert config["Keithley", "27XX", self.instr, "MODULE01", "module_name"] == cards[0], cards[0]
                    self.configured_modules["MODULE01"] = cards[0]
                except KeyError as err:
                    logger.error("{}: configuration {} does not exist.".format(KeyError, err))
                except AssertionError as err:
                    logger.error("{}: Switching module {} does not match any configuration".format(
                        AssertionError, str(err)))
                try:
                    assert config["Keithley", "27XX", self.instr, "MODULE02", "module_name"] == cards[1], cards[1]
                    self.configured_modules["MODULE02"] = cards[1]
                except KeyError as err:
                    logger.error("{}: configuration {} does not exist." .format(KeyError, err))
                except AssertionError as err:
                    logger.error("{}: Switching module {} does not match any configuration".format(
                        AssertionError, str(err)))
                logger.info("Configured modules : {}".format(self.configured_modules))
                try:
                    if config["Keithley", "27XX", self.instr, 'MODULE01', 'module_name']\
                            in self.non_amp_modules_list:
                        self.non_amp_module["MODULE01"] = True
                    if config["Keithley", "27XX", self.instr, 'MODULE02', 'module_name']\
                            in self.non_amp_modules_list:
                        self.non_amp_module["MODULE02"] = True
                except KeyError:
                    pass
                logger.info("Hardware initialized")
            except AttributeError:
                logger.error(AttributeError)
        except visa.errors.VisaIOError as err:
            logger.error(err)

    def configuration_sequence(self):
        """Configure each channel selected by the user

        Read the configuration file to get the channels used and their configuration,
        and send the keithley a sequence allowing to set up each channel.
        
        :raises TypeError: Channel section of configuration file not correctly defined, each channel should be a dict
        :raises ValueError: Channel not correctly defined, it should at least contain a key called "mode"
        """
        logger.info("       ********** CONFIGURATION SEQUENCE INITIALIZED **********")

        self.reset()
        self.clear_buffer()
        channels = ''

        # The following loop set up each channel in the config file
        for module in self.configured_modules:
            for key in config["Keithley", "27XX", self.instr, module, "CHANNELS"].keys():

                # Handling user mistakes if the channels' configuration section is not correctly set up
                if not type(config["Keithley", "27XX", self.instr, module, 'CHANNELS', key]) == dict:
                    logger.info("Channel {} not correctly defined, must be a dictionary" .format(key))
                    continue
                if not config["Keithley", "27XX", self.instr, module, 'CHANNELS', key]:
                    continue
                if "mode" not in config["Keithley", "27XX", self.instr, module, 'CHANNELS', key]:
                    logger.info("Channel {} not fully defined, 'mode' is missing" .format(key))
                    continue
                if config["Keithley", "27XX", self.instr, module, 'CHANNELS', key, "mode"].upper()\
                        not in self.modes_channels_dict.keys():
                    logger.info("Channel {} not correctly defined, mode not recognized" .format(key))
                    continue

                # Channel mode
                mode = config["Keithley", "27XX", self.instr, module, 'CHANNELS', key, "mode"].upper()
                self.modes_channels_dict[mode].append(int(key))
                channel = '(@' + key + ')'
                channels += key + ","
                cmd = "FUNC '" + mode + "'," + channel
                self._instr.write(cmd)

                # Config
                if 'range' in config["Keithley", "27XX", self.instr, module, 'CHANNELS', key].keys():
                    rang = config["Keithley", "27XX", self.instr, module, 'CHANNELS', key, "range"]
                    if 'autorange' in str(rang):
                        self._instr.write(mode + ':RANG:AUTO ')
                    else:
                        self._instr.write(mode + ':RANG ' + str(range))

                if 'resolution' in config["Keithley", "27XX", self.instr, module, 'CHANNELS', key].keys():
                    resolution = config["Keithley", "27XX", self.instr, module, 'CHANNELS', key, "resolution"]
                    self._instr.write(mode + ':DIG ' + str(resolution))

                if 'nplc' in config["Keithley", "27XX", self.instr, module, 'CHANNELS', key].keys():
                    nplc = config["Keithley", "27XX", self.instr, module, 'CHANNELS', key, "nplc"]
                    self._instr.write(mode + ':NPLC ' + str(nplc))

                if "TEMP" in mode:
                    transducer = config["Keithley", "27XX", self.instr, module, 'CHANNELS', key, "transducer"].upper()
                    if "TC" in transducer:
                        tc_type = config["Keithley", "27XX", self.instr, module, 'CHANNELS', key, "type"].upper()
                        ref_junc = config["Keithley", "27XX", self.instr, module, 'CHANNELS', key, "ref_junc"].upper()
                        self.mode_temp_tc(channel, transducer, tc_type, ref_junc)
                    elif "THER" in transducer:
                        ther_type = config["Keithley", "27XX", self.instr, module, 'CHANNELS', key, "type"].upper()
                        self.mode_temp_ther(channel, transducer, ther_type)
                    elif "FRTD" in transducer:
                        frtd_type = config["Keithley", "27XX", self.instr, module, 'CHANNELS', key, "type"].upper()
                        self.mode_temp_frtd(channel, transducer, frtd_type)

                # Console info
                logger.info("Channels {} \n {}".format(key,
                                                       config["Keithley", "27XX", self.instr, module, 'CHANNELS', key]))
                # Timeout update for long measurement modes such as voltage AC
                if "AC" in mode:
                    self._instr.timeout += 4000

                # Handling errors from Keithley
                current_error = self.get_error()
                try:
                    if current_error != '0,"No error"':
                        raise ValueError("The following error has been raised by the Keithley:\
                        %s => Please refer to the User Manual to correct it\n\
                        Note: To make sure channels are well configured in the .toml file,\
                        refer to section 15 'SCPI Reference Tables', Table 15-5" % current_error)
                except Exception as err:
                    logger.info("{}".format(err))
                    pass
        
        self.current_mode = 'scan_list'
        self.channels_scan_list = channels[:-1]
        logger.info("       ********** CONFIGURATION SEQUENCE SUCCESSFULLY ENDED **********")

    def clear_buffer(self):
        # Default: auto clear when scan start
        self._instr.write("TRAC:CLE")

    def clear_buffer_off(self):
        # Disable buffer auto clear
        self._instr.write("TRAC:CLE:AUTO OFF")

    def clear_buffer_on(self):
        # Disable buffer auto clear
        self._instr.write("TRAC:CLE:AUTO ON")

    def close(self):
        self._instr.write("ROUT:OPEN:ALL")
        self._instr.close()

    def data(self):
        """Get data from instrument

        Make the Keithley perform 3 actions: init, trigger, fetch. Then process the answer to return 3 variables:
        - The answer (string)
        - The measurement values (numpy array)
        - The timestamp of each measurement (numpy array)
        """
        if not self.sample_count_1:
            # Initiate scan
            self._instr.write("INIT")
            # Trigger scan
            self._instr.write("*TRG")
            # Get data (equivalent to TRAC:DATA? from buffer)
            str_answer = self._instr.query("FETCH?")
        else:
            str_answer = self._instr.query("FETCH?")
        # Split the instrument answer (MEASUREMENT,TIME,READING COUNT) to create a list
        list_split_answer = str_answer.split(",")

        # MEASUREMENT & TIME EXTRACTION
        list_measurements = list_split_answer[::3]
        str_measurements = ''
        list_times = list_split_answer[1::3]
        str_times = ''
        for j in range(len(list_measurements)):
            if not j == 0:
                str_measurements += ','
                str_times += ','
            for l1 in range(len(list_measurements[j])):
                test_carac = list_measurements[j][-(l1+1)]
                # Remove non-digit characters (units)
                if test_carac.isdigit():
                    if l1 == 0:
                        str_measurements += list_measurements[j]
                    else:
                        str_measurements += list_measurements[j][:-l1]
                    break
            for l2 in range(len(list_times[j])):
                test_carac = list_times[j][-(l2+1)]
                # Remove non-digit characters (units)
                if test_carac.isdigit():
                    if l2 == 0:
                        str_times += list_times[j]
                    else:
                        str_times += list_times[j][:-l2]
                    break

        # Split created string to access each value
        list_measurements_values = str_measurements.split(",")
        list_times_values = str_times.split(",")
        # Create numpy.array containing desired values (float type)
        array_measurements_values = np.array(list_measurements_values, dtype=float)
        if not self.sample_count_1:
            array_times_values = np.array(list_times_values, dtype=float)
        else:
            array_times_values = np.array([0], dtype=float)

        return str_answer, array_measurements_values, array_times_values

    def get_card(self):
        # Query switching module
        return self._instr.query("*OPT?")
    
    def get_error(self):
        # Ask the keithley to return the last current error
        return self._instr.query("SYST:ERR?")
    
    def get_idn(self):
        # Query identification
        return self._instr.query("*IDN?")
    
    def init_cont_off(self):
        # Disable continuous initiation
        self._instr.write("INIT:CONT OFF")
        
    def init_cont_on(self):
        # Enable continuous initiation
        self._instr.write("INIT:CONT ON")

    def mode_temp_frtd(self, channel, transducer, frtd_type,):
        self._instr.write("TEMP:TRAN " + transducer + "," + channel)
        self._instr.write("TEMP:FRTD:TYPE " + frtd_type + "," + channel)

    def mode_temp_tc(self, channel, transducer, tc_type, ref_junc,):
        self._instr.write("TEMP:TRAN " + transducer + "," + channel)
        self._instr.write("TEMP:TC:TYPE " + tc_type + "," + channel)
        self._instr.write("TEMP:RJUN:RSEL " + ref_junc + "," + channel)

    def mode_temp_ther(self, channel, transducer, ther_type,):
        self._instr.write("TEMP:TRAN " + transducer + "," + channel)
        self._instr.write("TEMP:THER:TYPE " + ther_type + "," + channel)
    
    def reset(self):
        # Clear measurement event register
        self._instr.write("*CLS")
        # One-shot measurement mode (Equivalent to INIT:COUNT OFF)
        self._instr.write("*RST")

    def set_mode(self, mode):
        """Define whether the Keithley will scan all the scan_list or only channels in the selected mode

        :param mode: Supported modes: 'SCAN_LIST', 'VDC', 'VAC', 'IDC', 'IAC', 'R2W', 'R4W', 'FREQ' and 'TEMP'
        :type mode: string
        """
        mode = mode.upper()
        
        # FRONT panel
        if "SCAN" not in mode:
            self.init_cont_on()
            self.sample_count_1 = True
            self.reading_scan_list = False
            self._instr.write("FUNC '" + mode + "'")

        # REAR panel
        else:
            self.clear_buffer()
            # Init continuous disabled
            self.init_cont_off()
            mode = mode[5:]
            self.current_mode = mode
            if 'SCAN_LIST' in mode:
                self.reading_scan_list = True
                self.sample_count_1 = False
                channels = '(@' + self.channels_scan_list + ')'
                # Set to perform 1 to INF scan(s)
                self._instr.write("TRIG:COUN 1")
                # Trigger immediately after previous scan end if IMM
                self._instr.write("TRIG:SOUR BUS")
                # Set to scan <n> channels
                samp_count = 1 + channels.count(',')
                self._instr.write("SAMP:COUN "+str(samp_count))
                # Disable scan if currently enabled
                self._instr.write("ROUT:SCAN:LSEL NONE")
                # Set scan list channels
                self._instr.write("ROUT:SCAN " + channels)
                # Start scan immediately when enabled and triggered
                self._instr.write("ROUT:SCAN:TSO IMM")
                # Enable scan
                self._instr.write("ROUT:SCAN:LSEL INT")

            else:
                self.reading_scan_list = False
                # Select channels in the channels list (config file) matching the requested mode
                channels = '(@' + str(self.modes_channels_dict[mode])[1:-1] + ')'
                # Set to perform 1 to INF scan(s)
                self._instr.write("TRIG:COUN 1")
                # Set to scan <n> channels
                samp_count = 1+channels.count(',')
                self._instr.write("SAMP:COUN "+str(samp_count))
                if samp_count == 1:
                    self.init_cont_on()
                    # Trigger definition
                    self._instr.write("TRIG:SOUR IMM")
                    # Disable scan if currently enabled
                    self._instr.write("ROUT:SCAN:LSEL NONE")
                    self._instr.write("ROUT:CLOS " + channels)
                    
                    self._instr.write("FUNC '" + mode + "'")
                    logger.info("rear sample count: {}".format(self.sample_count_1))
                    if not self.sample_count_1:
                        self.sample_count_1 = True
                    self.reading_scan_list = False
                else:
                    self.sample_count_1 = False
                    # Trigger definition
                    self._instr.write("TRIG:SOUR BUS")
                    # Disable scan if currently enabled
                    self._instr.write("ROUT:SCAN:LSEL NONE")
                    # Set scan list channels
                    self._instr.write("ROUT:SCAN " + channels)
                    # Start scan immediately when enabled and triggered
                    self._instr.write("ROUT:SCAN:TSO IMM")
                    # Enable scan
                    self._instr.write("ROUT:SCAN:LSEL INT")
                
            return channels
        
    def stop_acquisition(self):
        # If scan in process, stop it
        self._instr.write("ROUT:SCAN:LSEL NONE")

    def user_command(self):
        command = input('Enter here a command you want to send directly to the Keithley [if None, press enter]: ')
        if command != '':
            if command[-1] == "?":
                print(self._instr.query(command))
            else:
                self._instr.write(command)
            self.user_command()
        else:
            pass


if __name__ == "__main__":
    try:
        print("In main")

        # You can use this main section for:
        # - Testing connexion and communication with your instrument
        # - Testing new methods in developer mode

        RM = visa.ResourceManager("@ivi")
        print("list resources", RM.list_resources())

        # K2700 Instance of KeithleyVISADriver class (replace ASRL1::INSTR by the name of your resource)
        k2700 = Keithley27XXVISADriver("ASRL1::INSTR")
        k2700.init_hardware()
        print("IDN?")
        print(k2700.get_idn())
        k2700.reset()
        k2700.configuration_sequence()

        # Daq_viewer simulation first run
        k2700.set_mode(str(input('Enter which mode you want to scan \
        [scan_scan_list, scan_volt:dc, scan_r2w, scan_temp...]:')))
        print('Manual scan example of command set to send directly: >init >*trg >trac:data?')
        k2700.user_command()
        print('Automatic scan example with 2 iterations')
        for i in range(2):
            print(k2700.data())
        print(k2700.data())

        # Daq_viewer simulation change mode
        k2700.set_mode(str(input('Enter which mode you want to scan \
        [scan_scan_list, scan_volt:dc, scan_r2w, scan_temp...]:')))
        for i in range(2):
            print(k2700.data())
        print(k2700.data())

        k2700.clear_buffer()
        k2700.close()

        print("Out")

    except Exception as e:
        print("Exception ({}): {}".format(type(e), str(e)))
